"""
           Correspondence Analysis
    cf. Gifi,1991. Nonlinear Multivariate Analysis, 
           reprinted with correction
                 pp. 265-271

    The script was written by Y. Okamoto, 2022.12
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

#   ExcelファイルをPandasのDataFrame型として読み込む
fl_nm = input('Excel Data File(*.xlsx) = ')
xlfile = pd.ExcelFile(fl_nm)
df = pd.read_excel(xlfile)

#    １行目の値（列変数名）
col_names = df.keys()         #    Excelの１行目
c00 = col_names[0]            #    行変数名の列のキー名
col_names = col_names[1:]     #    ２番目以降を取り出す
print(col_names)

#    行変数名
row_names = df[c00]
print(row_names)

#    セル内の数値
df_data = df.values              #   ２行目以降が取り出される
F = df_data[:,1:]                #    取り出した各行の２列目以降を取り出す

r = F.sum(axis = 1)              #    行単位の和
r = np.reshape(r, (len(r), 1))   #    １次元配列から２次元配列（行列）への型変換
N = r.sum()                      #    データの総和
c = F.sum(axis = 0)              #    列単位の和
c = np.reshape(c, (len(c), 1))   #   １次元配列から２次元配列（行列）への型変換


E = r @ c.T / N                  #    Expected values
print('E =\n', E)

N_2_A = (F - E) / (E**0.5)             #    The last line in page 269..
print(N_2_A)
print('X^2 =', np.sum(N_2_A**2))

A = N_2_A / (N**0.5)                   #     eq.(8.24) 
print('A =\n', A)

#     Aの特異値分解SVD
A = np.asfarray(A)                                 #    Aの要素の型変換
K, Lmbd, Lh = np.linalg.svd(A, full_matrices = False)
L = Lh.T
p = sum([1 if v > 1.0e-7 else 0 for v in Lmbd])    #    布置の次元
print('p =', p)
Lmbd = Lmbd[:p]                          #    0でない特異値
print('K =\n', K)
K = K[:,:p]                              #    0でない特異値に対応する左特異ベクトル
print('K =\n', K)
print('Lmbd =', Lmbd)
print('Lmbd**2 =', Lmbd**2)
print('X^2 =', N * np.sum(Lmbd**2))
print('L =\n', L)     
L = L[:,:p]                              #  0でない特異値に対応する右特異ベクトル
print('L =\n', L)

X = ((N**0.5) * np.diag(np.reshape(r**-0.5, (1, len(r)))[0]))   \
      @ K @ np.diag(Lmbd)                 #   eq.(8.19)
print('X =\n', X)

Y = ((N**0.5) * np.diag(np.reshape(c**-0.5, (1, len(c)))[0])) @ L
print(Y)

XYp = X @ Y.T                            #   Projection of Y onto X
print(XYp)

x_axis = 1
y_axis = 2 if p > 1 else 1
while True:
    plt.plot(X.T[x_axis-1], X.T[y_axis-1], 'o', label = 'Rows')
    for i in range(len(row_names)):
        plt.text(X.T[x_axis-1][i]+0.01, X.T[y_axis-1][i]-0.01, row_names[i])
    plt.plot(Y.T[x_axis-1], Y.T[y_axis-1], 'o', label = 'Cols')
    for i in range(len(col_names)):
        plt.text(Y.T[x_axis-1][i]+0.01, Y.T[y_axis-1][i]-0.01, col_names[i])
    plt.legend()
    plt.xlabel('Dim-{0:}'.format(x_axis), fontsize = 14)
    plt.ylabel('Dim-{0:}'.format(y_axis), fontsize = 14)
    plt.savefig(f'Figdim{x_axis}dim{y_axis}.png')
    print(f'Figdim{x_axis}dim{y_axis}.png was saved.')
    plt.show()
    x_axis = int(input('X-axis(<={0:}; if <= 0, then quit.) = '.format(p)))
    if x_axis <= 0:
        break
    elif x_axis > p:
        x_axis = p
    y_axis = int(input('Y-axis(<={0:}; if <= 0, then quit.) = '.format(p)))
    if y_axis <= 0:
        break
    elif y_axis > p:
        y_axis = p
print('CA ended.')


